

unsigned long
host2ip (char *host)
{
  struct hostent *hp;
  struct hostent he;
  unsigned long addr;

  if (!(hp = gethostbyname (host)))
    {
//    perror("gethostbyname");
      return -1;
    }

  memcpy (&he, hp, sizeof (he));

  if (he.h_length != 4)
    return -1;

  memcpy (&addr, he.h_addr_list[0], 4);

  return addr;
}

int
connect_ip (unsigned long ip, unsigned short port, unsigned short sport)
{
  struct sockaddr_in lad, rad;
  int s;
  int sflag;
  fd_set fdset;
  struct timeval timeout;
  int opt, optlen;
  int delay;


  lad.sin_family = AF_INET;
  lad.sin_addr.s_addr = htonl (INADDR_ANY);
  lad.sin_port = htons (sport);

  rad.sin_family = AF_INET;
  rad.sin_addr.s_addr = ip;
  rad.sin_port = htons (port);

  if ((s = socket (AF_INET, SOCK_STREAM, IPPROTO_TCP)) == -1)
    {
      perror ("socket");
      return -1;
    }

  if (sport != 0)
    {
      if (bind (s, (struct sockaddr *) &lad, sizeof (lad)))
	{
	  perror ("bind");
	  close (s);
	  return -1;
	}
    }

  sflag = fcntl (s, F_GETFL, 0);
  fcntl (s, F_SETFL, sflag | O_NONBLOCK);

  if (connect (s, (struct sockaddr *) &rad, sizeof (rad)) != 0)
    {

      if (errno == EINPROGRESS)
	{
	  srand (time (NULL));
	  delay = 500000 + 50000 * ((int) (5.0 * rand () / (RAND_MAX + 1.0)));
//          fprintf(stderr,"connect_ip: usleep %i\n",delay);
	  usleep (delay);
	  FD_ZERO (&fdset);
	  FD_SET (s, &fdset);
	  timeout.tv_usec = 0;
	  timeout.tv_sec = REPLY_TIMEOUT;
	  if (select (s + 1, NULL, &fdset, NULL, &timeout) == 1)
	    {
	      optlen = sizeof (opt);
	      getsockopt (s, SOL_SOCKET, SO_ERROR, &opt, &optlen);
	      if (opt == 0)
		{
		  fcntl (s, F_SETFL, sflag);
		  return s;
		}
	      else
		{
		  fprintf (stderr, "connect_ip: getsockopt return %i\n", opt);
		  close (s);
		  return -1;
		}
	    }
	  else
	    {
	      perror ("select");
	      close (s);
	      return -1;
	    }
	}
      else
	{
	  perror ("connect");
	  close (s);
	  return -1;
	}
    }

  fcntl (s, F_SETFL, sflag);
  return s;
}



int
getanswer (int fd, int sec, char *reply, unsigned long size)
{
  fd_set fds;
  struct timeval max;
  struct timeval timeout;
  struct timeval now;
  unsigned long reply_pos;
  char ch;
  int nbread;


  gettimeofday (&max, NULL);
  max.tv_sec += sec;

  reply_pos = 0;
  while ((reply_pos + 1) < size)
    {
      FD_ZERO (&fds);
      FD_SET (fd, &fds);

      gettimeofday (&now, NULL);
      if (now.tv_usec > max.tv_usec)
	{
	  timeout.tv_usec = max.tv_usec + 1000000 - now.tv_usec;
	  timeout.tv_sec = max.tv_sec - now.tv_sec - 1;
	}
      else
	{
	  timeout.tv_usec = max.tv_usec - now.tv_usec;
	  timeout.tv_sec = max.tv_sec - now.tv_sec;
	}
      if ((timeout.tv_usec < 0) || (timeout.tv_sec < 0))
	return -1;

      if (select (fd + 1, &fds, NULL, NULL, &timeout) == -1)
	{
	  perror ("select");
	  break;
	}

      if (!FD_ISSET (fd, &fds))
	{
	  fprintf (stderr, "read timeout\n");
	  break;
	}

      if ((nbread = read (fd, &ch, 1)) == 0)
	{
	  break;
	}

      if (nbread != 1)
	{
	  perror ("read");
	  break;
	}

//      write(1,&ch,1);

      if (reply != NULL)
	reply[reply_pos] = ch;
      reply_pos++;
    }

  if (reply != NULL)
    reply[reply_pos] = 0;

  return reply_pos;
}





int
getreply (int fd, int sec, char *reply, unsigned long size)
{
  fd_set fds;
  struct timeval max;
  struct timeval timeout;
  struct timeval now;
  char msg[10];
  unsigned long pos;
  unsigned long reply_pos;
  char ch;
  int nbread;


  gettimeofday (&max, NULL);
  max.tv_sec += sec;

  reply_pos = 0;
  do
    {
      pos = 0;
      memset (msg, 0, sizeof (msg));
      while (1)
	{
	  FD_ZERO (&fds);
	  FD_SET (fd, &fds);

	  gettimeofday (&now, NULL);
	  if (now.tv_usec > max.tv_usec)
	    {
	      timeout.tv_usec = max.tv_usec + 1000000 - now.tv_usec;
	      timeout.tv_sec = max.tv_sec - now.tv_sec - 1;
	    }
	  else
	    {
	      timeout.tv_usec = max.tv_usec - now.tv_usec;
	      timeout.tv_sec = max.tv_sec - now.tv_sec;
	    }
	  if ((timeout.tv_usec < 0) || (timeout.tv_sec < 0))
	    return -1;

	  if (select (fd + 1, &fds, NULL, NULL, &timeout) == -1)
	    {
	      perror ("select");
	      return -1;
	    }

	  if (!FD_ISSET (fd, &fds))
	    {
	      fprintf (stderr, "read timeout\n");
	      return -1;
	    }

	  if ((nbread = read (fd, &ch, 1)) == 0)
	    {
	      break;
	    }

	  if (nbread != 1)
	    {
	      perror ("read");
	      break;
	    }

//      write(1,&ch,1);

	  if (pos < 4)
	    msg[pos] = ch;
	  if ((reply != NULL) && reply_pos < size)
	    reply[reply_pos] = ch;
	  pos++;
	  reply_pos++;

	  if (ch == '\n')
	    break;
	}

      if (reply != NULL)
	reply[reply_pos] = 0;

    }
  while ((pos < 3) || (((pos > 3) && (msg[3] == '-')) || (msg[0] == ' ')));


  return atoi (msg);
}



int
getendreply (int fd, int sec, char *reply, unsigned long size)
{
  int code;

  while (1)
    {
      code = getreply (fd, sec, reply, size);
      if ((code >= 200) || (code <= 0))
	break;
    }

  return code;
}



int
writes (int fd, char *str)
{
//  printf("write: %s\n",str);
  return (write (fd, str, strlen (str)));
}



int
ftploginpass (int ftpsock, char *login, char *passwd, int getbanner)
{
  char answer[1024];
  int reply;
  memset (answer, 0, sizeof (answer));
  if (getbanner == 0)
    {
      reply = getendreply (ftpsock, REPLY_TIMEOUT, answer, sizeof (answer));
      if (reply != 220)
	{
	  //  fprintf(stderr,"can't get ftp banner: %s\n",answer);
	  return -2;
	}
    }

  memset (answer, 0, sizeof (answer));
  snprintf (answer, sizeof (answer) - 10, "USER %s\r\n", login);
  if (writes (ftpsock, answer) < 0)
    {
//    printf ("WRITE SUCCK!\n");
      return (-2);
    }
  memset (answer, 0, sizeof (answer));

  reply = getendreply (ftpsock, REPLY_TIMEOUT, answer, sizeof (answer));
  if ((reply < 300) || (reply >= 400))
    {
      //  fprintf(stderr,"can't login (step USER): %s\n",answer);
      return -1;
    }


  memset (answer, 0, sizeof (answer));
  snprintf (answer, sizeof (answer) - 10, "PASS %s\r\n", passwd);
  if (writes (ftpsock, answer) < 0)
    {
      return (-2);
    }
  memset (answer, 0, sizeof (answer));

  reply = getendreply (ftpsock, REPLY_TIMEOUT, answer, sizeof (answer));
//  fprintf(stderr,"answer reply :%s\n",answer);
  if ((reply < 200) || (reply >= 300))
    {
//    fprintf(stderr,"can't login (step PASS): %s\n",answer);
      return -1;
    }

  return 1;
}
